/*  event.c

    Event file for  kview-rgb  (X11 image/movie display tool for Karma).

    Copyright (C) 1998  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable on-screen imaging of three arrayfiles of
    the general data structure format. The arrayfiles must contain a 2 or 3
    dimensional "Intelligent Array".
    This module runs on an X11 server.


    Written by      Richard Gooch   5-SEP-1998

    Last updated by Richard Gooch   6-SEP-1998: Remove "DATAMIN" and "DATAMAX"
  keywords from TrueColour image.


*/
#include <stdio.h>
#include <stdlib.h>
#include <karma.h>
#define NEW_WIN_SCALE
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include "kview-rgb.h"


/*  Private data  */
static multi_array *image_desc = NULL;
static unsigned int index = 0;
static ViewableImage vimage = NULL;
static iarray red_plane = NULL;
static iarray green_plane = NULL;
static iarray blue_plane = NULL;


/*  Public functions follow  */

flag process_set (Dataset *set, multi_array *multi_desc, KWorldCanvas canvas)
/*  [SUMMARY] Process a dataset.
    <set> The dataset.
    <multi_desc> The new data.
    <canvas> The canvas to display on.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    extern Dataset red_set, green_set, blue_set;
    static char *elem_names[3] = {"Red Intensity", "Green Intensity",
				  "Blue Intensity"};
    static char function_name[] = "process_set";

    if (multi_desc->num_arrays != 1)
    {
	fprintf (stderr, "Only one general data structure supported\n");
	return (FALSE);
    }
    if ( (image_desc != NULL) && (red_set.array == NULL) &&
	 (green_set.array == NULL) && (blue_set.array == NULL) )
    {
	ds_dealloc_multi (image_desc);
	image_desc = NULL;
	iarray_dealloc (red_plane);
	red_plane = NULL;
	iarray_dealloc (green_plane);
	green_plane = NULL;
	iarray_dealloc (blue_plane);
	blue_plane = NULL;
	viewimg_destroy (vimage);
	vimage = NULL;
    }
    if (set->array != NULL) iarray_dealloc (set->array);
    set->array = NULL;
    /*  Try to get 2-D image  */
    if ( !iarray_create_and_setup (&set->array, multi_desc, TRUE, 2, NONE,
				   &set->min, &set->max, FALSE, NULL) )
	return (FALSE);
    set->low_clip = set->min;
    set->high_clip = set->max;
    if (image_desc == NULL)
    {
	/*  Create a new one  */
	array_pointer arrayp;
	array_desc *arr_desc;
	multi_array *new_multi_desc = iarray_multi_array (set->array);
	packet_desc *top_pack_desc = new_multi_desc->headers[0];
	packet_desc *pack_desc;

	for (index = 0; index < top_pack_desc->num_elements; ++index)
	{
	    if (iarray_array_desc (set->array) ==
		(array_desc *) top_pack_desc->element_desc[index]) break;
	}
	
	/*  Copy descriptor  */
	if ( ( image_desc = ds_alloc_multi (1) ) == NULL )
	{
	    m_error_notify (function_name, "multi_array");
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	image_desc->headers[0] = ds_copy_desc_until (top_pack_desc, NULL);
	if (image_desc->headers[0] == NULL)
	{
	    m_error_notify (function_name, "multi_array");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	/*  Go in and frob the packet descriptor  */
	arr_desc = (array_desc *) image_desc->headers[0]->element_desc[index];
	ds_dealloc_packet (arr_desc->packet, NULL);
	if ( ( arr_desc->packet = ds_alloc_packet_desc (3) ) == NULL )
	{
	    m_error_notify (function_name, "packet desctriptor");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	pack_desc = arr_desc->packet;
	for (count = 0; count < 3; ++count)
	{
	    if ( ( pack_desc->element_desc[count] =
		   st_dup (elem_names[count]) ) == NULL )
	    {
		m_error_notify (function_name, "element name");
		ds_dealloc_multi (image_desc);
		image_desc = NULL;
		iarray_dealloc (set->array);
		set->array = NULL;
		return (FALSE);
	    }
	    pack_desc->element_types[count] = K_UBYTE;
	}
	/*  Allocate data  */
	image_desc->data[0] = ds_alloc_data (image_desc->headers[0],
					     TRUE, TRUE);
	if (image_desc->data[0] == NULL)
	{
	    m_error_notify (function_name, "element name");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	/*  Copy part data  */
	ds_copy_data (new_multi_desc->headers[0], new_multi_desc->data[0],
		      image_desc->headers[0], image_desc->data[0]);
	m_copy (&arrayp,
		image_desc->data[0] +
		ds_get_element_offset (image_desc->headers[0], index),
		sizeof arrayp);
	/*  Get rid of crud  */
	ds_remove_unique_named_element (image_desc->headers[0],
					image_desc->data, "DATAMIN", TRUE);
	ds_remove_unique_named_element (image_desc->headers[0],
					image_desc->data, "DATAMAX", TRUE);
	/*  Set up the colour plane intelligent arrays  */
	if ( ( red_plane =
	       iarray_get_from_multi_array (image_desc, NULL, 2, NULL,
					    elem_names[0]) ) == NULL )
	{
	    m_error_notify (function_name, "red iarray");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	if ( ( green_plane =
	       iarray_get_from_multi_array (image_desc, NULL, 2, NULL,
					    elem_names[1]) ) == NULL )
	{
	    m_error_notify (function_name, "green iarray");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (red_plane);
	    red_plane = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	if ( ( blue_plane =
	       iarray_get_from_multi_array (image_desc, NULL, 2, NULL,
					    elem_names[2]) ) == NULL )
	{
	    m_error_notify (function_name, "blue iarray");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (red_plane);
	    red_plane = NULL;
	    iarray_dealloc (green_plane);
	    green_plane = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	if ( ( vimage =
	       viewimg_create_rgb (canvas, image_desc, arr_desc, arrayp.array,
				   1, 0, 0, 1, 2, 0, NULL, NULL) ) == NULL )
	{
	    m_error_notify (function_name, "blue iarray");
	    ds_dealloc_multi (image_desc);
	    image_desc = NULL;
	    iarray_dealloc (red_plane);
	    red_plane = NULL;
	    iarray_dealloc (green_plane);
	    green_plane = NULL;
	    iarray_dealloc (blue_plane);
	    blue_plane = NULL;
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
    }
    else
    {
	/*  Verify it's the same size  */
	if ( iarray_dim_length (set->array, 1) !=
	     iarray_dim_length (red_plane, 1) )
	{
	    fprintf ( stderr, "X length: %lu doesn't match old length: %lu\n",
		      iarray_dim_length (set->array, 1),
		      iarray_dim_length (red_plane, 1) );
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
	if ( iarray_dim_length (set->array, 0) !=
	     iarray_dim_length (red_plane, 0) )
	{
	    fprintf ( stderr, "Y length: %lu doesn't match old length: %lu\n",
		      iarray_dim_length (set->array, 0),
		      iarray_dim_length (red_plane, 0) );
	    iarray_dealloc (set->array);
	    set->array = NULL;
	    return (FALSE);
	}
    }
    process_range (set);
    return (TRUE);
}   /*  End Function process_set  */

void process_range (Dataset *set)
/*  [SUMMARY] Process a dataset range.
    <set> The dataset.
    [RETURNS] Nothing.
*/
{
    iarray plane = NULL;
    double scale, offset;
    extern Dataset red_set, green_set, blue_set;

    if (set == &red_set) plane = red_plane;
    else if (set == &green_set) plane = green_plane;
    else if (set == &blue_set) plane = blue_plane;
    scale = 255.0 / (set->high_clip - set->low_clip);
    offset = -scale * set->low_clip;
    iarray_clip_scale_and_offset (plane, set->array, scale, offset,
				  set->low_clip, set->high_clip, FALSE);
    viewimg_register_data_change (vimage);
    viewimg_make_active (vimage);
}   /*  End Function process_range  */
