/*  main_xt.c

    Main file for  kview  (X11 image/movie display tool for Karma).

    Copyright (C) 1997-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable on-screen imaging of two arrayfiles of
    the general data structure format. The arrayfiles must contain a 2 or 3
    dimensional "Intelligent Array".
    This module runs on an X11 server.


    Written by      Richard Gooch   22-JAN-1997: Copied from kview/main_xt.c

    Updated by      Richard Gooch   2-FEB-1997: Added "filename" association to
  dataset(s).

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "Dataset1" and
  "Dataset2" events.

    Updated by      Richard Gooch   21-FEB-1997: Renamed to <kview> and deleted
  old <kview> and <koverlay>.

    Updated by      Richard Gooch   25-FEB-1997: Fixed bug when associating
  filename with dataset: check for PseudoColour or TrueColour dataset.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   6-JUN-1997: Support loading of datasets
  from command line.

    Updated by      Richard Gooch   18-JUN-1997: Removed IEEE floating point
  error handler setup.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   3-JUL-1997: Set foreground colour for
  DirectColour canvas to white to ensure white can always be allocated.

    Updated by      Richard Gooch   4-AUG-1997: Set resources for middle and
  right mouse button zooming.

    Updated by      Richard Gooch   10-AUG-1997: Created -create_at_init switch

    Updated by      Richard Gooch   12-AUG-1997: No longer initialise <contour>
  package because ImageDisplay widget does it.

    Updated by      Richard Gooch   19-AUG-1997: Set resources in create call
  for second fileselector.

    Updated by      Richard Gooch   3-OCT-1997: Strip leading "./" in
  command-line dataset names.

    Updated by      Richard Gooch   24-NOV-1997: Tried a few experimental
  hacks to make function pointer passing more C++ friendly.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu and bumped minor version number because panner implemented.

    Updated by      Richard Gooch   1-JAN-1998: Made use of new XkwNwmtitle
  resource.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   9-MAY-1998: No longer attach filename
  object.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   16-SEP-1998: Fixed bug in <data_event_func>
  where all datasets were captured.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   24-OCT-1998: Select on data name in
  <data_event_func>.

    Updated by      Richard Gooch   3-APR-1999: No longer set module name.

    Updated by      Richard Gooch   4-APR-1999: Only create second file button
  if module name is "kview".

    Updated by      Richard Gooch   10-JUN-1999: Made use of new function
  <xtmisc_init_post_realise_arg_process>.

    Updated by      Richard Gooch   5-JUL-1999: Added "-quiet" option.

    Updated by      Richard Gooch   2-AUG-1999: Supported unloading of data.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Last updated by Richard Gooch   3-OCT-2001: Added translations for Filewin.


*/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <karma.h>
#define NEW_WIN_SCALE
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_dir.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_ic.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Filewin.h>
#include <Xkw/View2Datasets.h>
#include "kview.h"

#define VERSION "1.1.18"


/*  Private functions  */
STATIC_FUNCTION (flag data_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


/*  Private data  */
String fallback_resources[] =
{
    "Kview*pseudoColourCanvas*background:              black",
    "Kview*directColourCanvas*background:              black",
    "Kview*trueColourCanvas*background:                black",
    "Kview*pseudoColourStereoCanvas*background:        black",
    "Kview*directColourStereoCanvas*background:        black",
    "Kview*trueColourStereoCanvas*background:          black",
    "Kview*AnimateControl*positionCanvas.background:   black",
    "Kview*pseudoColourCanvas*foreground:              white",
    "Kview*directColourCanvas*foreground:              white",
    "Kview*Command*background:                         grey70",
    "Kview*Repeater*background:                        grey70",
    "Kview*Ktoggle*background:                         grey80",
    "Kview*closeButton*background:                     grey90",
    "Kview*ChoiceMenu.background:                      turquoise",
    "Kview*ExclusiveMenu.background:                   turquoise",
    "Kview*Value*background:                           #d0a0a0",
    "Kview*ImageDisplay*quit*background:               orange",
    "Kview*SimpleSlider.foreground:                    Sea Green",
    "Kview*SimpleProgressMeter.background:             turquoise3",
    "Kview*SimpleProgressMeter.foreground:             grey60",
    "Kview*ImageDisplay*trackLabel0*font:              8x13bold",
    "Kview*ImageDisplay*trackLabel1*font:              8x13bold",
    "Kview*ImageDisplay*trackLabel2*font:              8x13bold",
    "Kview*SimpleMenu*font:                            10x20",
    "Kview*ImageDisplay.Command.font:                  7x13bold",
    "Kview*ImageDisplay.applicationBox.Command.font:   7x13bold",
    "Kview*ImageDisplay.Ktoggle.font:                  7x13bold",
    "Kview*ImageDisplay.ChoiceMenu.font:               7x13bold",
    "Kview*ImageDisplay.exportMenu.menuButton.font:    7x13bold",
    "Kview*ImageDisplay.overlayMenu.menuButton.font:   7x13bold",
    "Kview*ImageDisplay*zoomMenu*Unzoom*foreground:    red",
    "Kview*SimpleSlider.borderWidth:                   0",
    "Kview*IntensityPolicy*showIntensityReset:         True",
    "Kview*IntensityPolicy*resetIntensityToggle.state: True",
    "Kview*IntensityPolicy*autoIntensityScale:         False",
    "Kview*ZoomPolicy*showMiddleZoom:                  True",
    "Kview*ZoomPolicy*middleZoomToggle.state:          True",
    "Kview*ZoomPolicy*showRightZoom:                   True",
    "Kview*ZoomPolicy*rightZoomToggle.state:           True",
    "Kview*Filewin.dirname.textSink.font:              6x13",
    "Kview*Filewin*list.translations:            #override<MouseMoved>: Set()",
    "Kview*font:                                       9x15bold",
    "Kview*borderColor:                                black",
    "Kview*background:                                 aquamarine",
    "Kview*foreground:                                 black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-cmap_master", ".topForm.cmapMaster", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-create_at_init", "*createAtInit", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget image_display = NULL;
static Widget twinviewpopup = NULL;


int main (int argc, char **argv)
{
    KWorldCanvas wc_pseudo, wc_direct, wc_true;
    int argcount;
    XtAppContext app_context;
    Widget filewin, filepopup, box, w;
    extern flag quiet;
    extern char title_name[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    /*static char function_name[] = "main";*/

    /*  Initialise module  */
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kview",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 100,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    image_display = XtVaCreateManagedWidget ("topForm",
					     imageDisplayWidgetClass,
					     main_shell,
					     XtNborderWidth, 0,
					     XkwNview2Datasets, TRUE,
					     XkwNnumDatasets, 2,
					     XkwNnumTrackLabels, 3,
					     NULL);
    if ( ( w = XtNameToWidget (image_display, "filesButton") ) != NULL )
    {
	/*  Force main "Files" button to be re-labelled as "Set1"  */
	XtVaSetValues (w,
		       XtNlabel, "Set1",
		       NULL);
    }
    filepopup = XtNameToWidget (image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
		   XkwNdomainName, "FILE1",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNforwardSyntheticEvents, True,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    if (strcmp (module_name, "kview") == 0)
    {
	/*  Get the application box  */
	box = XtNameToWidget (image_display, "applicationBox");
	/*  Create the Set2 file selector and popup button  */
	filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
					  main_shell,
					  XtNtitle,
					  "Kview File Selector",
					  XtNx, 0,
					  XtNy, 0,
					  XkwNautoPopdown, True,
					  XkwNfilenameTester,
					  XkwFilewinStandardFileTester_nD,
					  XkwNdomainName, "FILE2",
					  XkwNloadFiles, TRUE,
					  NULL);
	w = XtVaCreateManagedWidget ("filesButton", commandWidgetClass, box,
				     XtNlabel, "Set2",
				     NULL);
	XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, filepopup);
	XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
	XtVaSetValues (filewin,
		       XkwNtrapDirectoryDatasets, True,
		       NULL);
    }
    ds_event_register_func (data_event_func, NULL);
    twinviewpopup = XtNameToWidget (image_display, "twinviewpopup");
    XtVaSetValues (twinviewpopup,
		   XkwNwmtitle, title_name,
		   NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &wc_pseudo,
		   XkwNdirectColourCanvas, &wc_direct,
		   XkwNtrueColourCanvas, &wc_true,
		   NULL);
    if (wc_pseudo != NULL)
    {
	viewimg_register_position_event_func (wc_pseudo,
					      XkwView2DatasetsCanvasEventHandler,
					      (void *) twinviewpopup);
    }
    if (wc_direct != NULL)
    {
	viewimg_register_position_event_func (wc_direct,
					      XkwView2DatasetsCanvasEventHandler,
					      (void *) twinviewpopup);
    }
    if (wc_true != NULL)
    {
	viewimg_register_position_event_func (wc_true,
					      XkwView2DatasetsCanvasEventHandler,
					      (void *) twinviewpopup);
    }
    xtmisc_init_post_realise_arg_process (&argc, argv);
    /*  Option processing after widgets created and realised  */
    for (argcount = 1; argcount < argc; ++argcount)
    {
	if ( (strcmp (argv[argcount], "-set1") == 0) ||
	     (strcmp (argv[argcount], "-set2") == 0) )
	{
	    int set;
	    multi_array *multi_desc;
	    char txt[STRING_LENGTH], fname[STRING_LENGTH];

	    set = atoi (argv[argcount] + 4);
	    if ( (set != 1) && (set != 2) )
	    {
		fprintf (stderr, "Option: \"%s\" ignored\n", argv[argcount]);
		continue;
	    }
	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "File missing for set%d\n", set);
		continue;
	    }
	    ++argcount;
	    sprintf (txt, "FILE%d", set);
	    if (strncmp (argv[argcount], "./", 2) == 0)
	    {
		strcpy (fname, argv[argcount] + 2);
	    }
	    else strcpy (fname, argv[argcount]);
	    if ( ( multi_desc =
		   foreign_guess_and_read (fname, K_CH_MAP_LOCAL,
					   FALSE, NULL,
					   FA_GUESS_READ_END) )
		 == NULL )
	    {
		fprintf (stderr, "Error reading file: \"%s\"\n", fname);
		continue;
	    }
	    ds_event_dispatch (multi_desc, txt, fname);
	    ds_dealloc_multi (multi_desc);
	    m_clear (fname, STRING_LENGTH);  /*  Force me to take care  */
	}
	else if (strcmp (argv[argcount], "-quiet") == 0) quiet = TRUE;
	else fprintf (stderr, "Option: \"%s\" ignored\n", argv[argcount]);
    }
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main  */


/*  Private functions follow  */

static flag data_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    Dataset *set = NULL;
    unsigned int set_number = 0;
    char *ptr;
    char filename[STRING_LENGTH], stripped_filename[STRING_LENGTH];
    extern Dataset set0, set1;

    if (strcmp (domain, "FILE1") == 0) set_number = 0;
    else if (strcmp (domain, "FILE2") == 0) set_number = 1;
    else
    {
	if (strcmp (name, "Dataset1") == 0) set_number = 0;
	else if (strcmp (name, "Dataset2") == 0) set_number = 1;
	else if (strncmp (name, "connection:", 11) == 0) set_number = 0;
	else return (FALSE);
    }
    set = (set_number == 0) ? &set0 : &set1;
    XkwView2DatasetsRegisterDataset (twinviewpopup, NULL, 0.0, 0.0,
				     NULL, NULL, NULL, set_number);
    if ( !process_set (multi_desc, set) ) return (FALSE);
    XkwView2DatasetsRegisterDataset (twinviewpopup,
				     set->array_pseudo, set->min, set->max,
				     set->array_red, set->array_green,
				     set->array_blue, set_number);
    if (strcmp (filename, "UNLOAD") == 0) return (TRUE);
    if (strncmp (domain, "FILE", 4) == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    XtVaSetValues (image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    return (TRUE);
}   /*  End Function data_event_func  */

#ifdef OS_SunOS

static void dummy ()
/*  [PURPOSE] This routine is needed for SunOS, because of a problem with
    shared archive use. The <<_XtToolkitInitialize>> symbol is defined in
    the Xt shared archive, but it appears the shared archive is not linked for
    this application. Other applications (like <<krenzo>>) have no such
    problems. This is a truly bizzare bug.
*/
{
    XtToolkitInitialize ();
}   /*  End Function dummy  */
#endif
